/* 
 *  /////////////////////////////////////////////////////////////
 *  
 *  4164/4864 DRAM Tester by TheDefpom - Rev 1.0 4th January 2022
 *  https://www.youtube.com/user/TheDefpom
 *  
 *  /////////////////////////////////////////////////////////////
 *  
 *  This is a heavily modified and improved version of the original code as referenced below.
 *  I added an OLED display, test button input, progress indication and information about connections.
 *  
 *  PORTD IS DIGITAL PINS 0-7 on the ATmega168 & 328P- these are used for the actual signal lines, BUT they clash with the serial TX/RX pins on 0 and 1, so be carefull how they are used.
 *
 *  //////////////////////////////
 *  CONNECTIONS:
 *  Connect the A0-A7 pins of the DRAM socket to the D0-D7 pins.
 *  
 *  You can change the below connections around if you need to, these dont matter as much, adjust them to suit your board layout.
 *  Connect DOUT to pin 8
 *  Connect CAS to pin 9
 *  Connect RAS to pin 10
 *  Connect WE to pin 11
 *  Connect DIN to pin 12
 *  Pin 13 is the onboard LED indicator, you can connect an external LED too if you wish, this shows when the IC has power turned on, and also blinks after each test is completed
 *  Connect the test button to pin 14
 *  Connect OLED I2C SDA to A4 (assuming a Pro Mini)
 *  Connect OLED I2C SCL to A5 (assuming a Pro Mini)
 *  //////////////////////////////
 *  
 *  Based upon the basic code by:
 *  Noel Llopis 2021 https://www.youtube.com/watch?v=0x9EOhDEzPw
 *  and Chris Osborn <fozztexx@fozztexx.com>
 *  http://insentricity.com/a.cl/252
 */

#define DIN   12
#define DOUT  8
#define CAS   9
#define RAS   10
#define WE    11
#define LED   13
#define TEST  14
#define POWERON 15

#define ADDR_BITS 8
#define COLROW_COUNT (1 << ADDR_BITS)

#include <Wire.h> // I2C support
#include <Adafruit_SSD1306.h> // OLED display support

#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels
#define OLED_RESET -1
#define OLED_ADDR 0x3C // OLED display TWI address
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// 128 x 64 pixel display
#if (SSD1306_LCDHEIGHT != 64)
#error("Height incorrect, please fix Adafruit_SSD1306.h!");
#endif


String line1text = ""; // line 1 text holder
String line2text = ""; // line 2 text holder
String line3text = ""; // line 3 text holder
byte line1textsize = 1; // line 1 text size default
byte line2textsize = 1; // line 2 text size default
byte line3textsize = 1; // line 3 text size default
 
byte failed = 0; // failed flag
byte lastcol = 0; // column count tracking for display updates



void setup(){
  Wire.begin(); // Standard SDA (A4), SCL (A5) pins
  
  pinMode(LED, OUTPUT);
  pinMode(TEST, INPUT_PULLUP);
  pinMode(POWERON, OUTPUT);
  
  pinMode(DIN, OUTPUT);
  pinMode(DOUT, INPUT);

  pinMode(CAS, OUTPUT);
  pinMode(RAS, OUTPUT);
  pinMode(WE, OUTPUT);

  DDRD = 0xff;

  // turn off all power to the IC so it is safe to unplug it and plug in a new one for another test
  readFromRowCol(0, 0); // turn off all data lines
  digitalWrite(CAS, LOW); // turn off column select
  digitalWrite(RAS, LOW); // turn off row select
  digitalWrite(WE, LOW); // turn off write enable
  digitalWrite(DIN, LOW); // turn off DATA In
  digitalWrite(POWERON, HIGH); // turn off chip power

  
  digitalWrite(LED, HIGH);

  // initialize and clear OLED display
  display.begin(SSD1306_SWITCHCAPVCC, OLED_ADDR);
  //display.setRotation(2); // works, kind of, offsets the text to the right though (it just flips the text in place)
  display.clearDisplay();
  display.setTextColor(WHITE); // set OLED display colour
  display.display();

  // show on display
  line1text = "4164/4864";
  line2text = "DRAM";
  line3text = "Tester";
  line1textsize = 2;
  line2textsize = 2;
  line3textsize = 2;
  doNewDisplay();
  delay(2000);
  
  // show on display
  line1text = "By Scott";
  line2text = "TheDefpom.com";
  line3text = "youtube.com/TheDefpom";
  line1textsize = 2;
  line2textsize = 1;
  line3textsize = 1;
  doNewDisplay();
  delay(2500);
  
  // show on display
  line1text = "Insert IC";
  line2text = " ";
  line3text = " ";
  line1textsize = 2;
  line2textsize = 2;
  line3textsize = 2;
  doNewDisplay();
  delay(50);  
}




// handle displaying of data on display
void doNewDisplay() {
  display.clearDisplay();
  // line 1
  if (line1text != "") { // only try to print text if there is something to do
    display.setCursor(0, 0);
    display.setTextSize(line1textsize);
    display.print(line1text);
  }
  // line 2
  if (line2text != "") { // only try to print text if there is something to do
    display.setCursor(00, 25);
    display.setTextSize(line2textsize);
    display.print(line2text);
  }
  // line 3
  if (line3text != "") { // only try to print text if there is something to do
    display.setCursor(0, 48);
    display.setTextSize(line3textsize);
    display.print(line3text);
  }
  display.display();
}


static inline void writeToRowCol(int row, int col){
  PORTD = row;
  digitalWrite(RAS, LOW);
  PORTD = col;
  digitalWrite(CAS, LOW);

  digitalWrite(WE, LOW);
  digitalWrite(WE, HIGH);
 
  digitalWrite(CAS, HIGH);
  digitalWrite(RAS, HIGH);
}


static inline int readFromRowCol(int row, int col){
  PORTD = row;
  digitalWrite(RAS, LOW);
  PORTD = col;
  digitalWrite(CAS, LOW);
  int val = digitalRead(DOUT); 
  digitalWrite(CAS, HIGH);
  digitalWrite(RAS, HIGH);
  return val;
}


void fail(int row, int col){
  digitalWrite(LED, LOW);
  //digitalWrite(FAILED, HIGH);

  // show on display
  line1text = "BAD IC";
  line2text = "Row:" + String(row);
  line3text = "Col:" + String(col);
  line1textsize = 2;
  line2textsize = 2;
  line3textsize = 2;
  doNewDisplay();
  delay(1);
  failed = 1;
}


// displays a progress indication, this will slow down the testing though!
void showWriting(){
  // show on display
  //line1text = "Testing...";
  //line2text = "";
  line3text = "Writing..";
  //line1textsize = 2;
  //line2textsize = 2;
  line3textsize = 2;
  doNewDisplay();
}

// displays a progress indication, this will slow down the testing though!
void showProgress(int col){
  // only update on every 10th column to speed up processing
  if(col >= lastcol + 5){
    // show on display
    //line1text = "Testing...";
    //line2text = "";
    line3text = "Col:" + String(col);
    //line1textsize = 2;
    //line2textsize = 2;
    line3textsize = 2;
    doNewDisplay();

    lastcol = col;
  }
}


void checkSame(int val){
  if(failed == 0){
    digitalWrite(DIN, val);

    // WRITE DATA
    showWriting(); // flag on display
    for (int col=0; col<COLROW_COUNT; col++){
      for (int row=0; row<COLROW_COUNT; row++){
        writeToRowCol(row, col);
      }
    }
  
    /* Reverse DIN in case DOUT is floating */
    digitalWrite(DIN, !val);
   
    // READ DATA
    for (int col=0; col<COLROW_COUNT; col++){
      // if it failed stop testing 
      if(failed == 1){
        break;
      }
      // progress indicator, disable to speed up testing
      showProgress(col);
      for (int row=0; row<COLROW_COUNT; row++){
        if (readFromRowCol(row, col) != val){
          fail(row,col);
          // if it failed stop testing 
          if(failed == 1){
            break;
          }
        }
      }
    }
  }
  lastcol = 0; // reset column tracking for display
  return;
}


void checkAlternating(int start){
  if(failed == 0){
    int i = start;
    // WRITE DATA
    showWriting(); // flag on display
    for (int col=0; col<COLROW_COUNT; col++){
      for (int row=0; row<COLROW_COUNT; row++){
        digitalWrite(DIN, i);
        i = !i;
        writeToRowCol(row, col);
      }
    }
    
    // READ DATA
    for (int col=0; col<COLROW_COUNT; col++){
      // if it failed stop testing 
      if(failed == 1){
        break;
      }
      // progress indicator, disable to speed up testing
      showProgress(col);
      for (int row=0; row<COLROW_COUNT; row++){ 
        if (readFromRowCol(row, col) != i)
          fail(row,col);
          // if it failed stop testing    
          if(failed == 1){
            break;
          }
        i = !i;
      }
    }
  }
  lastcol = 0; // reset column tracking for display
  return;
}


void checkRandom(){
  if(failed == 0){
    // Generate a somewhat random seed
    const int seed = analogRead(16);
  
    randomSeed(seed);
    // WRITE DATA
    showWriting(); // flag on display
    for (int col=0; col<COLROW_COUNT; col++){
      for (int row=0; row<COLROW_COUNT; row++){
        const int value = (int)random(2);    
        digitalWrite(DIN, value);
        writeToRowCol(row, col);
      }
    }
  
    // READ DATA
    // Set the same seed as for the write to know what value to get
    randomSeed(seed);
    for (int col=0; col<COLROW_COUNT; col++){
      // if it failed stop testing 
      if(failed == 1){
        break;
      }
      // progress indicator, disable to speed up testing
      showProgress(col);
      for (int row=0; row<COLROW_COUNT; row++){
        const int value = (int)random(2);    
        if (readFromRowCol(row, col) != value)
          fail(row,col);
          // if it failed stop testing 
          if(failed == 1){
            break;
          }
      }
    }
  }
  lastcol = 0; // reset column tracking for display
  return;
}




void loop(){
  // only run if button has been pressed, to avoid locking up the serial ports preventing the arduino from being re-programmed
  if(digitalRead(TEST) == LOW){
    failed = 0; // reset flag
    lastcol = 0; // reset column tracking for display
    
    digitalWrite(LED, HIGH);

    // turn on and initialise IC before trying to use it
    digitalWrite(POWERON, LOW); // turn on power to the chip  
    digitalWrite(CAS, HIGH);
    digitalWrite(RAS, HIGH);
    digitalWrite(WE, HIGH);
    delay(10); // this delay is needed to allow the chip time to power up before the tests start
    for (int i=0; i<10; i++){ // this is needed to wake up the IC, just reads some data before doing the tests, as specified in the IC data sheet.
      readFromRowCol(i, 0);
    }
    
    // show on display
    line1text = "Testing...";
    line2text = "Zeros";
    line3text = " ";
    line1textsize = 2;
    line2textsize = 2;
    line3textsize = 2;
    doNewDisplay();
    delay(50);
  
    digitalWrite(LED, LOW);
    delay(250);
    digitalWrite(LED, HIGH);
    
    
    checkSame(0);
  
    // if it didnt fail then do next test
    if(failed == 0){
      // show on display
      line1text = "Testing...";
      line2text = "Ones";
      line3text = " ";
      line1textsize = 2;
      line2textsize = 2;
      line3textsize = 2;
      doNewDisplay();      
      delay(10);
      
      digitalWrite(LED, LOW);
      delay(250);
      digitalWrite(LED, HIGH);
      
      checkSame(1);
    }
    
    // if it didnt fail then do next test
    if(failed == 0){
      // show on display
      line1text = "Testing...";
      line2text = "Alt. Zeros";
      line3text = " ";
      line1textsize = 2;
      line2textsize = 2;
      line3textsize = 2;
      doNewDisplay();
      delay(10);
      
      digitalWrite(LED, LOW);
      delay(250);
      digitalWrite(LED, HIGH);
      
      checkAlternating(0);
    }
    
    // if it didnt fail then do next test
    if(failed == 0){
      // show on display
      line1text = "Testing...";
      line2text = "Alt. Ones";
      line3text = " ";
      line1textsize = 2;
      line2textsize = 2;
      line3textsize = 2;
      doNewDisplay();
      delay(10);
      
      digitalWrite(LED, LOW);
      delay(250);
      digitalWrite(LED, HIGH);
      
      checkAlternating(1);
    }

    
    // if it didnt fail then do next test
    if(failed == 0){
      // show on display
      line1text = "Testing...";
      line2text = "Random";
      line3text = " ";
      line1textsize = 2;
      line2textsize = 2;
      line3textsize = 2;
      doNewDisplay();
      delay(10);
      
      digitalWrite(LED, LOW);
      delay(250);
      digitalWrite(LED, HIGH);
      
      checkRandom();      
    }
    
    // if it didnt fail then its probably OK
    if(failed == 0){
        // progress indicator, disable to speed up testing
      // show on display
      line1text = "Testing...";
      line2text = "PASSED!";
      line3text = " ";
      line1textsize = 2;
      line2textsize = 2;
      line3textsize = 2;
      doNewDisplay();
      delay(10);
    }

    // turn off all power to the IC so it is safe to unplug it and plug in a new one for another test
    readFromRowCol(0, 0); // turn off all data lines
    digitalWrite(CAS, LOW); // turn off column select
    digitalWrite(RAS, LOW); // turn off row select
    digitalWrite(WE, LOW); // turn off write enable
    digitalWrite(DIN, LOW); // turn off DATA In
    digitalWrite(POWERON, HIGH); // turn off chip power
    digitalWrite(LED, LOW);
  }
}
